Program DEM2TGA;

{
  Convert USGS data from Digital Elevation Model format to Targa
  image format. Based on DEM2POV.C ("modified by W. D. Kirby") and
  DEM2TGA.C by Jon Larimer.  This was created so there would be a
  better option for us DOS users than C programs!

  This outputs a 24-bit TGA.  Can be used for POV-type heightfields
  or for regular brighter=higher images.

  Please read DEM2TGA.TXT for more information.

  Traditional DISCLAIMER, etc:

  Do not redistribute this code unless the documentation (DEM2TGA.TXT)
  and the original executable (DEM2TGA.EXE) are included with it.  Do
  not modify this code in any way and redistribute it.  Do not
  distribute this with any other software or in any CD-ROM packages
  without the author's permission.  The author is not liable in any
  way for the function or dysfunction of this program on the user's
  computer.

  Now that that's over with, feel free to write if you find this program
  useful or interesting or inefficient or slow.  Bug complaints and
  comments of any kind are always welcome!

  Have fun,

  - Sean O'Malley, April 1999
    ffrog@geocities.com
    http://www.geocities.com/~ffrog
}

{ Input/output checking off.
  Extended syntax on - results of functions can be discarded.
  Debugging information off. }
{$I-,X+,D-}

{ diskbufferk: Size in kb, of a memory buffer to use.  Don't use less
  than 5 (kb).  blockread() and blockwrite() are really slow so this
  tries to use them minimally. }
const diskbufferk = 25;
      diskbuffersize = 512 * diskbufferk;

type inbuffertype  = array[1..diskbuffersize] of char;
     outbuffertype = array[1..diskbuffersize] of byte;

var inbuffer : ^inbuffertype;
    outbuffer : ^outbuffertype;
    color_range : word;
    elev_type : string[10];
    bufflocation, buffsize, outbufflocation : word;
    infile,outfile : file;
    infilename,outfilename : string[80];
    maxelev,minelev,rows,columns,elevrange : integer;
    colormap_scale : real;
    i,r,c : integer;
    cur_r,cur_c : integer;
    time : real;

function seconds:real;
var timer : longint absolute $40:$6c;
begin
 seconds := timer / 18.206;
end;

procedure fillbuff;
begin
 blockread(infile,inbuffer^,diskbuffersize,buffsize);
 if (buffsize = 0) then begin
  writeln('Fatal error: Unexpected end of file or error reading file.');
  halt(1);
 end;
 bufflocation := 1;
end;

function getchar:char;
begin
 getchar := inbuffer^[bufflocation];
 inc(bufflocation);
 if (bufflocation >= buffsize) then fillbuff;
end;

procedure dumpbyte(w:byte);
begin
 outbuffer^[outbufflocation] := w;
 inc(outbufflocation);
 if (outbufflocation > diskbuffersize) then begin
  blockwrite(outfile,outbuffer^,diskbuffersize);
  outbufflocation := 1;
 end;
end;

procedure flushbuffer;
begin
 if (outbufflocation > 1) then
  blockwrite(outfile,outbuffer^,(outbufflocation-1));
end;

function getstring(l:byte):string;
var outstring : string;
    loop : byte;
begin
 outstring[0] := chr(l);
 loop := 1;
 for loop := 1 to l do outstring[loop] := getchar;
 if (ioresult <> 0) then begin
  writeln('File read error.');
  halt(1);
 end;
 getstring := outstring;
end;

{ This is heavily modified based on DEM2POV rather than DEM2TGA
  for compatibility reasons, among other things. }
procedure writetgaheader(r,c:integer);
var i : byte;
    section : array[1..6] of byte;
    temp : shortint;
const null : byte = 0;
begin
 i := 0;
 while (i < 12) do begin
  if (i = 2) then dumpbyte(i) else dumpbyte(null);
  inc(i);
 end;
 section[1] := lo(c);   { Columns (X) }
 section[2] := hi(c);
 section[3] := lo(r);   { Rows    (Y) }
 section[4] := hi(r);
 section[5] := 24;      { *NOT* 8 bpp as in DEM2TGA! }
 section[6] := 32;      { "Bottom-down raster." }
 for i := 1 to 6 do dumpbyte(section[i]);
 { Notice no palette. :}
end;

function rltrim(trimstring:string):string;
{ Trim off white space on both ends. }
var add:byte;
    resstring:char;
begin
 resstring:=trimstring[1];
 add:=0;
 while resstring=' ' do begin
  inc(add);
  resstring:=trimstring[add+1];
 end;
 if add>0 then begin
  trimstring:=copy(trimstring,add+1,length(trimstring));
  add:=0;
 end;
 resstring:=trimstring[length(trimstring)];
 while resstring=' ' do begin
  inc(add);
  resstring:=trimstring[length(trimstring)-add];
 end;
 if add>0 then
  trimstring:=copy(trimstring,1,length(trimstring)-add);
 rltrim:=trimstring;
end;

{ Convert a string to lowercase. }
function lcase(f:string):string; assembler;
asm
 push ds
 les si,f
 lds di,@result
 xor ch,ch
 mov cl,es:[si]
 mov ds:[di],cl
 jcxz @ender
 inc si
 inc di
@loop1:
 mov al,es:[si]
 cmp al,'A'
 jb @skip
 cmp al,'Z'
 ja @skip
 add al,32
@skip:
 mov ds:[di],al
 inc di
 inc si
 loop @loop1
@ender:
 pop ds
end;

{ Search for searchchar in searchstring.  Return 0 if unsuccessful,
  otherwise return location. }
function charsearch(searchstring:string;searchchar:char):byte; assembler;
var origdi:word;
asm
 cld
 les di,searchstring
 mov cl,es:[di]
 xor ch,ch
 jcxz @badend
 mov al,searchchar
 inc di
 mov origdi,di
 repne scasb    
 jne @badend
 sub di,origdi
 mov ax,di
 jmp @end
@badend:
 xor al,al
@end:
end;

function nextfloat:real;
var ins:string;
    i:byte;
    code:integer;
    returner:real;
begin
 repeat
  ins[1] := getchar;
 until (ins[1] <> ' ');
 i := 2;
 repeat
  ins[i] := getchar;
  if (ins[i] = 'D') then ins[i] := 'E';
  inc(i);
 until (ins[i-1] = ' ');
 ins[0] := chr(i-2);
 val(ins,returner,code);
 nextfloat := returner;
end;

function nextint:integer;
var ins:string;
    code:integer;
    returner:integer;
    i:byte;
begin
 i := 1;
 repeat
  ins[1] := getchar;
 until (ins[1] <> ' ');
 repeat
  inc(i);
  ins[i] := getchar;
 until (ins[i] = ' ');
 ins[0] := chr(i-1);
 val(ins,returner,code);
 nextint := returner;
end;

procedure outtotga(what:integer);
var i : byte;
const triplet : array[1..3] of byte = (0,0,0);
begin
 { Blue, green, red.  24 bits.  Blue could be used below a certain
   threshold to simulate water. }
 { triplet[1] := 0; }
 triplet[2] := lo(what);
 triplet[3] := hi(what);
 for i := 1 to 3 do dumpbyte(triplet[i]);
end;

begin
 writeln('                         -=-');
 writeln('              DEM2TGA - Pascal Version');
 writeln('  Convert USGS .DEM data to .TGA image file format.');
 writeln('                         -=-');
 writeln('          This program based on C versions.');
 writeln('Sean O''Malley, March 1999 - E-mail: ffrog@geocities.com');
 writeln;
 if ((paramcount < 2) or (paramcount > 3)) then begin
  writeln('Usage:  DEM2TGA input_file output_file [-P]');
  writeln(' Where input_file is the DEM file to convert and output_file is the');
  writeln(' TGA file to create.  Extensions are optional; .DEM and .TGA assumed.');
  writeln(' Use -P to produce a POV (raytracer) heightfield image.');
  writeln;
  writeln('Example: dem2tga pullman pullman -p');
  writeln;
  writeln(' This will load PULLMAN.DEM and create a heightfield-type image');
  writeln(' called PULLMAN.TGA');
  halt(0);
 end;
 getmem(inbuffer,diskbuffersize);
 getmem(outbuffer,diskbuffersize);
 if ((inbuffer = nil) or (outbuffer = nil)) then begin
  writeln('Inadequate memory.');
  halt(1);
 end;
 infilename := lcase(paramstr(1));
 outfilename := lcase(paramstr(2));
 color_range := 255;
 if (paramcount = 3) then
  if (lcase(copy(paramstr(3),2,1)) = 'p') then color_range := 65535;
 if (charsearch(infilename,'.') = 0) then infilename := infilename + '.dem';
 if (charsearch(outfilename,'.') = 0) then outfilename := outfilename + '.tga';
 assign(infile,infilename);
 assign(outfile,outfilename);
 reset(infile,1);
 if (ioresult <> 0) then begin
  writeln('Unable to open file: ',infilename);
  halt(1);
 end;
 time := seconds;
 fillbuff;
 outbufflocation := 1;
 writeln('Quad name field: ',rltrim(getstring(144)));
 for i := 1 to 19 do nextint;
 i := nextint;
 write('Units code (ground planametric coordinates): ',i,' ');
 case i of
  0 : writeln('(radians)');
  1 : writeln('(feet)');
  2 : writeln('(meters)');
  3 : writeln('(arc-seconds)');
  else writeln('(unknown code)');
 end;
 i := nextint;
 write('Units code (elevation coordinates): ',i,' (');
 case i of
  1 : elev_type := 'feet';
  2 : elev_type := 'meters';
  else elev_type := '(unknown units)';
 end;
 writeln(elev_type,')');
 nextint;
 writeln('Ground coordinates of 4 corners of DEM:');
 for i := 1 to 4 do writeln('  ',nextfloat:1:3,'  ',nextfloat:1:3);
 writeln;
 minelev := trunc(nextfloat);
 maxelev := trunc(nextfloat);
 writeln('Minimum elevation: ',minelev,' ',elev_type);
 writeln('Maximum elevation: ',maxelev,' ',elev_type);
 elevrange := maxelev - minelev;
 writeln('Elevation range: ',elevrange,' ',elev_type);
 colormap_scale := color_range / elevrange;
 writeln('Color scaling value: ',colormap_scale:1:3);
 if (color_range = 255) then
  writeln('Creating standard TGA image.')
  else
  writeln('Creating POV-type heightfield image.');
 { Actually "don't need the next" *2* items!!! }
 nextint; nextint;
 cur_r := nextint;
 columns := nextint;
 cur_c := 1;
 writeln;
 write('Writing image ... ');
 rewrite(outfile,1);
 if (ioresult <> 0) then begin
  writeln('Error creating file: ',outfilename);
  halt(1);
 end;
 cur_r := nextint;
 cur_c := nextint;
 rows := nextint;
 writetgaheader(rows-1,columns);
 while (cur_c < columns) do begin
  nextint; nextint; nextint; nextint; nextint; nextint;
  for i := 1 to rows do
   outtotga(round((nextint - minelev) * colormap_scale));
  cur_r := nextint;
  cur_c := nextint;
  rows := nextint;
 end;
 writeln('Done.');
 time := seconds - time;
 writeln('Total time: ',time:1:3,' seconds');
 flushbuffer;
 close(infile);
 close(outfile);
 freemem(inbuffer,diskbuffersize);
 freemem(outbuffer,diskbuffersize);
end.
